import numpy as np

def calculate(w = 0.48, h = 0.2, df = 1.75, adjust = 1):
    """
    Calculates the flow multiplier factor, using the rounded rectangle model.

    ARGS:
    w: raster width (default 0.48mm) (float)
    h: raster height (default 0.2mm) (float)
    df: filament diameter (default 1.75mm) (float)
    adjust: adjust factor (default 100%) (float)

    RETURNS:
    Flow multiplier factor
    """
    a = 4 * w * h + (np.pi - 4) * h**2
    b = np.pi * df**2
    flow = adjust * a / b
    return flow


def extrude(x, y, flow):
    """
    Generates the extrusion coordinate array.

    ARGS:
    x: x array (array)
    y: y array (array)
    flow: flow multiplier (float)
    RETURNS:
    Extrusion coordinate array (array)
    """
    extrusion = np.zeros(len(x))
    for i in range(1, len(x)):
        dx = abs(x[i] - x[i - 1])
        dy = abs(y[i] - y[i - 1])
        extrusion[i] = np.sqrt((dx**2) + (dy**2)) * flow + extrusion[i-1]
    return extrusion



from altprint.path.contour import contour
from shapely.geometry import LineString
from shapely import affinity


class Layer:
    def __init__(self, **kwargs):
        prop_defaults = {
            "border": None,
            "height": 0.2,
            "gap": 0.5,
            "perimeters_num": 2,
            "perimeters": [],
            "infill": [],
            "skirt_num": [3, 5],
            "skirt": [],
        }

        for (prop, default) in prop_defaults.items():
            setattr(self, prop, kwargs.get(prop, default))

        for i in range(self.perimeters_num):
            self.perimeters.append(contour(self.border, - self.gap * i))

        for i in range(self.skirt_num[0]):
            self.skirt.append(contour(self.border, self.skirt_num[1] - self.gap * i))

        self.inner_border = contour(self.border,
                                    - self.gap * self.perimeters_num)

    def translate(self, x, y):
        self.border = affinity.translate(self.border, x, y)

        for i in range(self.perimeters_num):
            self.perimeters[i] = affinity.translate(self.perimeters[i], x, y)

        for i in range(len(self.infill)):
            self.infill[i] = affinity.translate(self.infill[i], x, y)

        self.inner_border = affinity.translate(self.inner_border, x, y)



from shapely.ops import split

def retract(x, y, f):
    """
    Splits an AB segment from a C point, between A and B. Used for creating
    an recovery zone on variable flow bridges.
    ARGS:
    x: list contaning x points from the segment (list)
    y: list contaning y points from the segment (list)
    f: factor that determines where the segment will be splited. (float)
    RETURNS:
    Tuple containg the two new segments. (tuple)
    """
    A = [x[0], y[0]]
    B = [x[1], y[1]]
    C = [A[0] + f*(B[0]-A[0]), A[1] + f*(B[1]-A[1])]
    acx = [A[0], C[0]]
    acy = [A[1], C[1]]
    cbx = [C[0], B[0]]
    cby = [C[1], B[1]]
    return acx, acy, cbx, cby

def split_lines(lines, spliter):
    final = []
    for line in lines:
        for i in list(split(line, spliter)):
            final.append(i)
    return final



from collections import deque
from shapely import affinity
from shapely.geometry import LineString
import numpy as np

def contour(border, L):
    """
    Generate a contour line from L milimeters of the shape object.

    ARGS:
    border: LineString of the border (LineString)
    L: distance of the contour from border. Positive numbers means external
    contours, negative numbers means internal contours (float)

    RETURNS:
    LineString containing the contour. (LineString)
    """
    border = border.coords[:-1]
    shape_border = []
    for i in range(len(border)):
        A = border[i-2]
        B = border[i-1]
        C = border[i]

        u = LineString([A,B])
        v = LineString([B,C])

        ux = B[0] - A[0]
        uy = B[1] - A[1]
        vx = C[0] - B[0]
        vy = C[1] - B[1]

        udx = L*uy/u.length
        udy = L*ux/u.length
        vdx = L*vy/v.length
        vdy = L*vx/v.length

        ut = affinity.translate(u,  udx, -udy)
        vt = affinity.translate(v, vdx, -vdy)
        Ax = ut.coords[0][0]
        Ay = ut.coords[0][1]
        Bx = ut.coords[1][0]
        By = ut.coords[1][1]
        Cx = vt.coords[0][0]
        Cy = vt.coords[0][1]
        Dx = vt.coords[1][0]
        Dy = vt.coords[1][1]

        a = np.array([[Ay-By, Bx-Ax], [Cy-Dy, Dx-Cx]])
        b = np.array([Bx*Ay-Ax*By,  Dx*Cy-Cx*Dy])
        x = np.linalg.solve(a, b)

        shape_border.append(x)
    shape_border = deque(shape_border)
    shape_border.rotate(-1)
    shape_border = list(shape_border)
    shape_border.append(shape_border[0])
    return LineString(shape_border)
